@@ -3,9 +3,13 @@ import socket
 import struct
 import types
 import io
+import re
 
 from pyroute2.common import hexdump
 
+_letters = re.compile('[A-Za-z]')
+_fmt_letters = re.compile('[^!><@=][!><@=]')
+
 ##  Netlink family
 #
 NETLINK_ROUTE = 0            # Routing/device hook
@@ -101,23 +105,11 @@ class nlmsg_base(dict):
         + NLA
         + ...
 
-    To describe data, you can use one of the ways:
+    To describe data, use python struct notation:
 
-    1. 'fmt' and 'fields' class atributes, e.g.:
+        fields = (('length', 'H'),
+                  ('type', 'H'))
 
-        fmt = 'HH'
-        fields = ('length',
-                  'type')
-
-        this will decode/encode two 16bit unsigned integers
-        into/from fields 'length' and 'type'.
-
-    2. 't_fields' attribute, e.g.:
-
-        t_fields = (('length', 'H'),
-                    ('type', 'H'))
-
-        will do the same.
 
     NLAs are decoded/encoded according to 'nla_map':
 
@@ -136,16 +128,19 @@ class nlmsg_base(dict):
     you can read in nlmsg_atoms class.
     '''
 
-    fmt = ''                    # data format string, see struct
-    fields = ('value', )        # data field names, to build a dictionary
-    t_fields = NotInitialized
-    header = None               # optional header class
-    nla_map = {}                # NLA mapping
+    fields = []                  # data field names, to build a dictionary
+    header = None                # optional header class
+    nla_map = {}                 # NLA mapping
 
     def __init__(self, buf=None, length=None, parent=None):
         dict.__init__(self)
         for i in self.fields:
-            self[i] = 0  # FIXME: only for number values
+            self[i[0]] = 0  # FIXME: only for number values
+        if isinstance(buf, basestring):
+            b = io.BytesIO()
+            b.write(buf)
+            b.seek(0)
+            buf = b
         self.buf = buf or io.BytesIO()
         self.length = length or 0
         self.parent = parent
@@ -156,9 +151,18 @@ class nlmsg_base(dict):
         self.value = NotInitialized
         if self.header is not None:
             self['header'] = self.header(self.buf)
-        self.register_fields()
         self.register_nlas()
 
+    def __eq__(self, value):
+        '''
+        Having nla, we are able to use it in operations like::
+
+            if nla == 'some value':
+                ...
+        '''
+        return self.getvalue() == value
+
+    @classmethod
     def nla2name(self, name):
         '''
         Convert NLA name into human-friendly name
@@ -169,6 +173,7 @@ class nlmsg_base(dict):
         '''
         return name[(name.find(self.prefix) + 1) * len(self.prefix):].lower()
 
+    @classmethod
     def name2nla(self, name):
         '''
         Convert human-friendly name into NLA name
@@ -188,7 +193,10 @@ class nlmsg_base(dict):
         to skip encoding of the header until some fields will
         be known.
         '''
-        self.buf.seek(struct.calcsize(self.fmt), 1)
+        size = 0
+        for i in self.fields:
+            size += struct.calcsize(i[1])
+        self.buf.seek(size, 1)
 
     def decode(self):
         self.offset = self.buf.tell()
@@ -199,17 +207,39 @@ class nlmsg_base(dict):
                 # update length from header
                 # it can not be less than 4
                 self.length = max(self['header']['length'], 4)
-                if self.fmt == 's':
-                    self.fmt = '%is' % (self.length - 4)
-                elif self.fmt == 'z':
-                    self.fmt = '%is' % (self.length - 5)
             except Exception as e:
                 raise NetlinkHeaderDecodeError(e)
         # decode the data
         try:
-            size = struct.calcsize(self.fmt)
-            self.update(dict(zip(self.fields,
-                             struct.unpack(self.fmt, self.buf.read(size)))))
+            for i in self.fields:
+                name = i[0]
+                fmt = i[1]
+
+                # 's' and 'z' can be used only in connection with
+                # length, encoded in the header
+                if fmt == 's':
+                    fmt = '%is' % (self.length - 4)
+                elif fmt == 'z':
+                    fmt = '%is' % (self.length - 5)
+
+                size = struct.calcsize(fmt)
+                raw = self.buf.read(size)
+                actual_size = len(raw)
+
+                # FIXME: adjust string size again
+                if fmt[-1] == 's':
+                    size = actual_size
+                    fmt = '%is' % (actual_size)
+                if size == actual_size:
+                    value = struct.unpack(fmt, raw)
+                    if len(value) == 1:
+                        self[name] = value[0]
+                    else:
+                        self[name] = value
+                else:
+                    # FIXME: log an error
+                    pass
+
         except Exception as e:
             raise NetlinkDataDecodeError(e)
         # decode NLA
@@ -217,7 +247,8 @@ class nlmsg_base(dict):
             # align NLA chain start
             self.buf.seek(NLMSG_ALIGN(self.buf.tell()))
             # read NLA chain
-            self.decode_nlas()
+            if self.nla_map:
+                self.decode_nlas()
         except Exception as e:
             raise NetlinkNLADecodeError(e)
         if len(self['attrs']) == 0:
@@ -227,28 +258,46 @@ class nlmsg_base(dict):
 
     def encode(self):
         init = self.buf.tell()
+        diff = 0
         # reserve space for the header
         if self.header is not None:
             self['header'].reserve()
-        if self.fmt == 's':
-            length = len(self.fields[0]) + 4
-            self.fmt = '%is' % (length)
-        elif self.fmt == 'z':
-            length = len(self.fields[0]) + 5
-            self.fmt = '%is' % (length)
-        payload = struct.pack(self.fmt, *([self[i] for i in self.fields]))
-        diff = NLMSG_ALIGN(len(payload)) - len(payload)
-        self.buf.write(payload)
-        self.buf.write(b'\0' * diff)
+
+        if self.getvalue() is not None:
+            try:
+                payload = b''
+                for i in self.fields:
+                    name = i[0]
+                    fmt = i[1]
+
+                    if fmt == 's':
+                        length = len(self[name])
+                        fmt = '%is' % (length)
+                    elif fmt == 'z':
+                        length = len(self[name]) + 1
+                        fmt = '%is' % (length)
+
+                    payload += struct.pack(fmt, self[name])
+
+            except Exception as e:
+                raise e
+            diff = NLMSG_ALIGN(len(payload)) - len(payload)
+            self.buf.write(payload)
+            self.buf.write(b'\0' * diff)
         # write NLA chain
-        self.encode_nlas()
+        if self.nla_map:
+            diff = 0
+            self.encode_nlas()
         # calculate the size and write it
         if self.header is not None:
-            save = self.buf.tell()
-            self['header']['length'] = save - init
-            self.buf.seek(init)
-            self['header'].encode()
-            self.buf.seek(save)
+            self.update_length(init, diff)
+
+    def update_length(self, start, diff):
+        save = self.buf.tell()
+        self['header']['length'] = save - start - diff
+        self.buf.seek(start)
+        self['header'].encode()
+        self.buf.seek(save)
 
     def setvalue(self, value):
         if type(value) is dict:
@@ -271,29 +320,12 @@ class nlmsg_base(dict):
         if self.value != NotInitialized:
             # value decoded by custom decoder
             return self.value
-        elif 'value' in self:
+
+        if 'value' in self and self['value'] != NotInitialized:
             # raw value got by generic decoder
             return self['value']
-        else:
-            return self
 
-    def register_fields(self):
-        '''
-        Convert 't_fields' attribute into 'fmt' and 'fields'.
-        '''
-        if self.t_fields is NotInitialized:
-            return
-
-        fields = []
-        fmt = []
-
-        for i in self.t_fields:
-            fmt.append(i[1])
-            if (i[1].find('x') == -1) and (i[0] != '__pad'):
-                fields.append(i[0])
-
-        self.fmt = ''.join(fmt)
-        self.fields = tuple(fields)
+        return self
 
     def register_nlas(self):
         '''
@@ -324,12 +356,17 @@ class nlmsg_base(dict):
             if i[0] in self.r_nla_map:
                 msg_class = self.r_nla_map[i[0]][0]
                 msg_type = self.r_nla_map[i[0]][1]
+                # is it a class or a function?
+                if isinstance(msg_class, types.MethodType):
+                    # if it is a function -- use it to get the class
+                    msg_class = msg_class()
                 try:
                     # encode NLA
                     nla = msg_class(self.buf, parent=self)
                     nla['header']['type'] = msg_type
                     nla.setvalue(i[1])
                     nla.encode()
+                    i[1] = nla
                 except:
                     # FIXME
                     import traceback
@@ -352,7 +389,7 @@ class nlmsg_base(dict):
                 # is it a class or a function?
                 if isinstance(msg_class, types.MethodType):
                     # if it is a function -- use it to get the class
-                    msg_class = msg_class()
+                    msg_class = msg_class(buf=self.buf, length=length)
                 # and the name
                 msg_name = self.t_nla_map[msg_type][1]
                 try:
@@ -373,8 +410,8 @@ class nlmsg_base(dict):
 
 
 class nla_header(nlmsg_base):
-    fmt = 'HH'
-    fields = ('length', 'type')
+    fields = (('length', 'H'),
+              ('type', 'H'))
 
 
 class nla_base(nlmsg_base):
@@ -385,8 +422,11 @@ class nlmsg_header(nlmsg_base):
     '''
     Common netlink message header
     '''
-    fmt = 'IHHII'
-    fields = ('length', 'type', 'flags', 'sequence_number', 'pid')
+    fields = (('length', 'I'),
+              ('type', 'H'),
+              ('flags', 'H'),
+              ('sequence_number', 'I'),
+              ('pid', 'I'))
 
 
 class nlmsg_atoms(nlmsg_base):
@@ -403,13 +443,13 @@ class nlmsg_atoms(nlmsg_base):
             self.value = None
 
     class uint8(nla_base):
-        fmt = '=B'
+        fields = [('value', 'B')]
 
     class uint16(nla_base):
-        fmt = '=H'
+        fields = [('value', 'H')]
 
     class uint32(nla_base):
-        fmt = '=I'
+        fields = [('value', 'I')]
 
     class ipaddr(nla_base):
         '''
@@ -420,7 +460,7 @@ class nlmsg_atoms(nlmsg_base):
         We do not specify here the string size, it will be
         calculated in runtime.
         '''
-        fmt = 's'
+        fields = [('value', 's')]
 
         def encode(self):
             self['value'] = socket.inet_pton(self.parent['family'], self.value)
@@ -434,7 +474,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
         Decode MAC address.
         '''
-        fmt = '=6s'
+        fields = [('value', '=6s')]
 
         def encode(self):
             self['value'] = ''.join((chr(int(i, 16)) for i in
@@ -449,7 +489,7 @@ class nlmsg_atoms(nlmsg_base):
         '''
         Represent NLA's content with header as hex string.
         '''
-        fmt = 's'
+        fields = [('value', 's')]
 
         def decode(self):
             nla_base.decode(self)
@@ -460,7 +500,7 @@ class nlmsg_atoms(nlmsg_base):
         Zero-terminated string.
         '''
         # FIXME: move z-string hacks from general decode here?
-        fmt = 'z'
+        fields = [('value', 'z')]
 
 
 class nla(nla_base, nlmsg_atoms):
@@ -483,10 +523,9 @@ class genlmsg(nlmsg):
     '''
     Generic netlink message
     '''
-    fmt = 'BBH'
-    fields = ('cmd',
-              'version',
-              'reserved')
+    fields = (('cmd', 'B'),
+              ('version', 'B'),
+              ('reserved', 'H'))
 
 
 class ctrlmsg(genlmsg):
@@ -501,4 +540,5 @@ class ctrlmsg(genlmsg):
                ('CTRL_ATTR_HDRSIZE', 'hex'),
                ('CTRL_ATTR_MAXATTR', 'hex'),
                ('CTRL_ATTR_OPS', 'hex'),
-               ('CTRL_ATTR_MCAST_GROUPS', 'hex'))
+               ('CTRL_ATTR_MCAST_GROUPS', 'hex'),
+               ('IPR_ATTR_SECRET', 'asciiz'))
