@@ -13,14 +13,15 @@ Quick start:
     ip['bala'].up()
 '''
 import re
-import copy
+import uuid
 import threading
 from socket import AF_INET
 from socket import AF_INET6
+from pyroute2.netlink import netlink_error
 from pyroute2.netlink.iproute import iproute
 from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
 
-nla_fields = [i[0] for i in ifinfmsg.nla_map]
+nla_fields = [ifinfmsg.nla2name(i[0]) for i in ifinfmsg.nla_map]
 nla_fields.append('flags')
 nla_fields.append('mask')
 nla_fields.append('change')
@@ -46,49 +47,41 @@ def get_addr_nla(msg):
     return nla
 
 
-class upset(set):
+class ipset(set):
 
-    def __init__(self):
-        set.__init__(self)
-        self.cleanup()
-        self.values = {}
+    def __init__(self, *argv, **kwarg):
+        set.__init__(self, *argv, **kwarg)
+        self.raw = {}
+        self.links = []
 
-    def commit(self):
-        for i in self.removed:
-            del self.values[i]
-        self.cleanup()
-
-    def cleanup(self):
-        self.added = set()
-        self.removed = set()
-
-    def pop(self, track=True):
-        item = set.pop(self)
-        if track:
-            self.removed.add(item)
-
-    def add(self, key, value=None, track=True):
-        if track:
-            self.added.add(key)
-        self.values[key] = value
+    def add(self, key, raw=None):
+        self.raw[key] = raw
         set.add(self, key)
+        for link in self.links:
+            link.add(key, raw)
 
-    def remove(self, key, value=None, track=True):
-        if track:
-            self.removed.add(key)
-        # do not remove the value: it will be kept for rollback
+    def remove(self, key, raw=None):
+        for link in self.links:
+            if key in link:
+                link.remove(key)
         set.remove(self, key)
 
+    def connect(self, link):
+        assert isinstance(link, ipset)
+        self.links.append(link)
+
 
 class dotkeys(dict):
 
     def __dir__(self):
         return [i for i in self if type(i) == str and _var_name.match(i)]
 
-    def __getattribute__(self, key):
+    def __getattribute__(self, key, *argv):
         try:
             return dict.__getattribute__(self, key)
-        except AttributeError:
+        except AttributeError as e:
+            if key == '__deepcopy__':
+                raise e
             return self[key]
 
     def __setattr__(self, key, value):
@@ -102,6 +95,94 @@ class dotkeys(dict):
             del self[key]
 
 
+class rwState(object):
+
+    def __init__(self):
+        self._rlock = threading.Lock()
+        self._wlock = threading.Lock()
+        self._lock = threading.Lock()
+        self._readers = 0
+        self._no_readers = threading.Event()
+        self._no_readers.set()
+        self._state = False
+
+    def __enter__(self):
+        self.reader_acquire()
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        self.reader_release()
+
+    def reader_acquire(self):
+        with self._wlock:
+            with self._rlock:
+                self._no_readers.clear()
+                self._readers += 1
+
+    def reader_release(self):
+        with self._rlock:
+            assert self._readers > 0
+            self._readers -= 1
+            if self._readers == 0:
+                self._no_readers.set()
+
+    def is_set(self):
+        return self._state
+
+    def acquire(self):
+        self._lock.acquire()
+        self._no_readers.wait()
+        self._state = True
+
+    def release(self):
+        with self._wlock:
+            self._no_readers.wait()
+            self._state = False
+            self._lock.release()
+
+
+class IPDBError(Exception):
+    pass
+
+
+class IPDBTransactionRequired(IPDBError):
+    pass
+
+
+def update(f):
+    def decorated(self, *argv, **kwarg):
+        # obtain update lock
+        tid = None
+        with self._snapshot_lock:
+            # fix the pass-through mode:
+            with self._direct_state:
+                kwarg['direct'] = self._direct_state.is_set()
+                if not kwarg['direct']:
+                    # 1. begin transaction for 'direct' type
+                    if self._mode == 'direct':
+                        tid = self.begin()
+                    # 2. begin transaction, if there is none
+                    elif self._mode == 'implicit':
+                        if not self._tids:
+                            self.begin()
+                    # 3. require open transaction for 'explicit' type
+                    elif self._mode == 'explicit':
+                        if not self._tids:
+                            raise IPDBTransactionRequired()
+                    # 4. transactions can not require transactions :)
+                    elif self._mode == 'snapshot':
+                        kwarg['direct'] = True
+                    # do not support other modes
+                    else:
+                        raise IPDBError('transaction mode not supported')
+                    # now that the transaction _is_ open
+                f(self, *argv, **kwarg)
+        if tid:
+            # close the transaction for 'direct' type
+            self.commit(tid)
+    return decorated
+
+
 class interface(dotkeys):
     '''
     Objects of this class represent network interface and
@@ -123,7 +204,7 @@ class interface(dotkeys):
     exception will be raised. Failed transaction review
     will be attached to the exception.
     '''
-    def __init__(self, dev, ipr=None, parent=None):
+    def __init__(self, dev=None, ipr=None, parent=None, mode='direct'):
         '''
         One can use interface objects standalone as
         well as in connection with ipdb object. Standalone
@@ -135,6 +216,7 @@ class interface(dotkeys):
             * parent -- ipdb() reference
         '''
         self.ip = ipr
+        self.uid = uuid.uuid4()
         self.cleanup = ('header',
                         'linkinfo',
                         'af_spec',
@@ -143,18 +225,64 @@ class interface(dotkeys):
                         'map',
                         'stats',
                         'stats64')
-        dev.prefix = 'IFLA_'
         self.last_error = None
+        self._direct_state = rwState()
         self._parent = parent
-        self._fields = [dev.nla2name(i) for i in nla_fields]
-        self._updated = {}
+        self._tids = []
+        self._transactions = {}
+        self._mode = mode
         self._slaves = dotkeys()
-        self.load(dev)
+        self._snapshot_lock = threading.RLock()
+        # local setup: direct state is required
+        self._direct_state.acquire()
+        self['ipaddr'] = ipset()
+        self._direct_state.release()
+        # 8<-----------------------------------
+        if dev is not None:
+            self.load(dev)
+
+    def pick(self, detached=True):
+        '''
+        Get a snapshot of the interface state. Can be of two
+        types:
+            * detached=True -- (default) "true" snapshot
+            * detached=False -- keep ip addr set updated from OS
+
+        Please note, that "updated" doesn't mean "in sync".
+        The reason behind this logic is that snapshots can be
+        used as transactions.
+        '''
+        with self._snapshot_lock:
+            res = interface(ipr=self.ip, mode='snapshot')
+            for key in tuple(self.keys()):
+                if key in nla_fields:
+                    res[key] = self[key]
+            res['ipaddr'] = ipset(self['ipaddr'])
+            if self._parent is not None and not detached:
+                self['ipaddr'].connect(res['ipaddr'])
+            return res
+
+    def __sub__(self, pif):
+        '''
+        '''
+        res = interface(ipr=self.ip, mode='snapshot')
+        # simple keys
+        for key in self:
+            if (key in nla_fields) and \
+                    ((key not in pif) or (self[key] != pif[key])):
+                res[key] = self[key]
+        # ip addresses
+        ipaddr = ipset(self['ipaddr'] - pif['ipaddr'])
+        if ipaddr:
+            res['ipaddr'] = ipaddr
+
+        return res
 
     def __hash__(self):
         return self['index']
 
     def __enter__(self):
+        self.begin()
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
@@ -185,17 +313,17 @@ class interface(dotkeys):
     def load(self, dev):
         '''
         Update the interface info from RTM_NEWLINK message.
+
+        This call always bypasses open transactions, loading
+        changes directly into the interface data.
         '''
-        dev.prefix = 'IFLA_'
+        self._direct_state.acquire()
         self.update(dev)
         self._attrs = set()
         for (name, value) in dev['attrs']:
-            norm = dev.nla2name(name)
+            norm = ifinfmsg.nla2name(name)
             self._attrs.add(norm)
             self[norm] = value
-        # save ifname for transaction
-        self.old_name = self['ifname']
-        self._updated = {}
         # load interface kind
         linkinfo = dev.get_attr('IFLA_LINKINFO')
         if linkinfo:
@@ -208,90 +336,181 @@ class interface(dotkeys):
         # the rest is possible only when interface
         # is used in IPDB, not standalone
         if self._parent is not None:
-            # load IP addresses from IPDB
+            # connect IP address set from IPDB
             self['ipaddr'] = self._parent.ipaddr[self['index']]
         # finally, cleanup all not needed
         for item in self.cleanup:
             if item in self:
                 del self[item]
+        self._direct_state.release()
 
-    def __setitem__(self, key, value):
-        if key in self:
-            self._updated[key] = self[key]
-        dotkeys.__setitem__(self, key, value)
+    @update
+    def __setitem__(self, key, value, direct):
+        if not direct:
+            transaction = self.last()
+            transaction[key] = value
+        else:
+            dotkeys.__setitem__(self, key, value)
+
+    @update
+    def __delitem__(self, key, direct):
+        if not direct:
+            transaction = self.last()
+            if key in transaction:
+                del transaction[key]
+        else:
+            dotkeys.__delitem__(self, key)
+
+    @update
+    def add_ip(self, ip, mask, direct):
+        if not direct:
+            transaction = self.last()
+            transaction.add_ip(ip, mask)
+        else:
+            self['ipaddr'].add((ip, mask))
+
+    @update
+    def del_ip(self, ip, mask, direct):
+        if not direct:
+            transaction = self.last()
+            if (ip, mask) in transaction['ipaddr']:
+                transaction.del_ip(ip, mask)
+        else:
+            self['ipaddr'].remove((ip, mask))
+
+    def begin(self):
+        '''
+        Start new transaction
+        '''
+        # keep snapshot's ip addr set updated from the OS
+        # it is required by the commit logic
+        t = self.pick(detached=False)
+        self._transactions[t.uid] = t
+        self._tids.append(t.uid)
+        return t.uid
+
+    def last(self):
+        '''
+        Return last open transaction
+        '''
+        if not self._tids:
+            raise IPDBTransactionRequired()
+        return self._transactions[self._tids[-1]]
 
     def review(self):
         '''
-        Review the changes that are not commited yet. Output
-        format can be changed later.
+        Review last open transaction
         '''
-        response = {'ipaddr': [],
-                    'attrs': {}}
-        for i in self['ipaddr'].added:
-            response['ipaddr'].append('+%s/%s' % tuple(i))
-        for i in self['ipaddr'].removed:
-            response['ipaddr'].append('-%s/%s' % tuple(i))
-        for i in tuple(self._updated.keys()):
-            response['attrs'][i] = '%s -> %s' % (self._updated[i],
-                                                 self[i])
-        return response
+        if not self._tids:
+            raise IPDBTransactionRequired()
+        added = self.last() - self
+        removed = self - self.last()
+        added['-ipaddr'] = removed['ipaddr']
+        added['+ipaddr'] = added['ipaddr']
+        del added['ipaddr']
+        return added
 
-    def rollback(self):
+    def drop(self):
         '''
         Drop all not applied changes and rollback transaction.
         '''
-        for i in self['ipaddr'].added:
-            self['ipaddr'].remove(i, track=False)
-        for i in self['ipaddr'].removed:
-            self['ipaddr'].add(i, track=False)
-        self['ipaddr'].cleanup()
-        for i in tuple(self._updated.keys()):
-            self[i] = self._updated[i]
-            del self._updated[i]
+        del self._transactions[self._tids.pop()]
 
-    def commit(self):
+    def reload(self):
+        '''
+        Reload interface information
+        '''
+        self._parent._addr_event.clear()
+        self['ipaddr'].clear()
+        self.ip.get_links(self['index'])
+        self.ip.get_addr()
+        self._parent._addr_event.wait()
+
+    def commit(self, tid=None, transaction=None, rollback=False):
         '''
         Commit transaction. In the case of exception all
         changes applied during commit will be reverted.
         '''
         e = None
-        transaction = copy.deepcopy(self.review())
+        added = None
+        removed = None
+        snapshot = self.pick()
+        if tid:
+            transaction = self._transactions[tid]
+        else:
+            transaction = transaction or self.last()
         try:
             # commit IP address changes
-            for i in self['ipaddr'].added:
-                # add address
+            removed = self - transaction
+            for i in removed['ipaddr']:
+                # When you remove a primary IP addr, all subnetwork
+                # can be removed. In this case you will fail, but
+                # it is OK, no need to roll back
+                try:
+                    self.ip.addr_del(self['index'], i[0], i[1])
+                except netlink_error as x:
+                    # bypass only errno 99, 'Cannot assign address'
+                    if x.errno != 99:
+                        raise x
+
+            added = transaction - self
+            for i in added['ipaddr']:
                 self.ip.addr_add(self['index'], i[0], i[1])
-            for i in self['ipaddr'].removed:
-                self.ip.addr_del(self['index'], i[0], i[1])
+
             # commit interface changes
             request = {}
-            for key in self._updated:
-                if key in self._fields:
-                    request[key] = self[key]
-            self.ip.link('set', self['index'], **request)
-            # flush IP address changes
-            self['ipaddr'].commit()
+            for key in added:
+                if key in nla_fields:
+                    request[key] = added[key]
+
+            # apply changes only if there is something to apply
+            if request:
+                self.ip.link('set', self['index'], **request)
+
         except Exception as e:
             # something went wrong: roll the transaction back
-            try:
-                # remove all added and add all removed
-                for i in self['ipaddr'].added:
-                    self.ip.addr_del(self['index'], i[0], i[1])
-                for i in self['ipaddr'].removed:
-                    self.ip.addr_add(self['index'], i[0], i[1])
-                # apply old attributes
-                request = {}
-                for key in tuple(self._updated.keys()):
-                    if key in self._fields:
-                        request[key] = self._updated[key]
-                        del self._updated[key]
-                self.ip.link('set', self['index'], **request)
-                # rollback IP address tracking
-                self['ipaddr'].cleanup()
-            except:
-                pass
-        # re-load interface information
-        self.load(self.ip.get_links(self['index'])[0])
+            if not rollback:
+                self.reload()
+                # 8<-----------------------------------------
+                # That's a hack, but we have to use it, since
+                # OS response can be not so fast
+                # * inject added IPs directly into self
+                # * wipe removed IPs from the interface data
+                #
+                # This info will be used to properly roll back
+                # changes.
+                #
+                # Still, there is a possibility that the
+                # rollback will run even before IP addrs will
+                # be assigned. But we can not cope with that.
+                self._direct_state.acquire()
+                if removed:
+                    for i in removed['ipaddr']:
+                        try:
+                            self['ipaddr'].remove(i)
+                        except KeyError:
+                            pass
+                if added:
+                    for i in added['ipaddr']:
+                        self['ipaddr'].add(i)
+                self._direct_state.release()
+                # 8<-----------------------------------------
+                self.commit(transaction=snapshot, rollback=True)
+            else:
+                # somethig went wrong during automatic rollback.
+                # that's the worst case, but it is still possible,
+                # since we have no locks on OS level.
+                self.drop()
+                raise e
+
+        # if it is not a rollback turn
+        if not rollback:
+            # drop last transaction in any case
+            self.drop()
+            # re-load interface information
+            self.reload()
+
+        # raise exception for failed transaction
         if e is not None:
             e.transaction = transaction
             raise e
@@ -322,7 +541,7 @@ class ipdb(dotkeys):
     No methods of the class should be called directly.
     '''
 
-    def __init__(self, ipr=None, host='localsystem',
+    def __init__(self, ipr=None, host='localsystem', mode='implicit',
                  key=None, cert=None, ca=None):
         '''
         Parameters:
@@ -335,6 +554,8 @@ class ipdb(dotkeys):
         only the first one will receive anything.
         '''
         self.ip = ipr or iproute(host=host, key=key, cert=cert, ca=ca)
+        self.mode = mode
+        self._stop = False
 
         # resolvers
         self.by_name = dotkeys()
@@ -344,6 +565,10 @@ class ipdb(dotkeys):
         self.ipaddr = {}
         self.routes = {}
         self.neighbors = {}
+        self.old_names = {}
+
+        # update events
+        self._addr_event = threading.Event()
 
         # load information on startup
         links = self.ip.get_links()
@@ -364,19 +589,25 @@ class ipdb(dotkeys):
         ret.append('by_index')
         return ret
 
+    def shutdown(self):
+        self._stop = True
+        self.ip.get_links()
+        self.ip.shutdown_sockets()
+
     def update_links(self, links):
         '''
         Rebuild links index from list of RTM_NEWLINK messages.
         '''
         for dev in links:
             if dev['index'] not in self.ipaddr:
-                self.ipaddr[dev['index']] = upset()
+                self.ipaddr[dev['index']] = ipset()
             i = \
                 self.by_index[dev['index']] = \
                 self[dev['index']] = \
-                interface(dev, self.ip, self)
+                interface(dev, self.ip, self, mode=self.mode)
             self[i['ifname']] = \
                 self.by_name[i['ifname']] = i
+            self.old_names[dev['index']] = i['ifname']
 
     def update_slaves(self, links):
         '''
@@ -409,10 +640,10 @@ class ipdb(dotkeys):
             if nla is not None:
                 method = getattr(self.ipaddr[addr['index']], action)
                 try:
-                    method(key=(nla, addr['prefixlen']),
-                           value=addr, track=False)
+                    method(key=(nla, addr['prefixlen']), raw=addr)
                 except:
                     pass
+        self._addr_event.set()
 
     def monitor(self):
         '''
@@ -420,7 +651,7 @@ class ipdb(dotkeys):
         default iproute queue and updates objects in the
         database.
         '''
-        while True:
+        while not self._stop:
             try:
                 messages = self.ip.get()
             except:
@@ -430,8 +661,8 @@ class ipdb(dotkeys):
                     index = msg['index']
                     if index in self:
                         # get old name
-                        old = self[index].old_name
-                        # load message
+                        old = self.old_names[index]
+                        # load interface from the message
                         self[index].load(msg)
                         # check for new name
                         if self[index]['ifname'] != old:
@@ -441,8 +672,11 @@ class ipdb(dotkeys):
                                 del self[old].if_master.if_slaves[old]
                             del self[old]
                             del self.by_name[old]
+                            if index in self.old_names:
+                                del self.old_names[index]
                             self[self[index]['ifname']] = self[index]
                             self.by_name[self[index]['ifname']] = self[index]
+                            self.old_names[index] = self[index]['ifname']
                     else:
                         self.update_links([msg])
                     self.update_slaves([msg])
@@ -452,6 +686,7 @@ class ipdb(dotkeys):
                         # FIXME catch exception
                         del self.by_name[self[msg['index']]['ifname']]
                         del self.by_index[msg['index']]
+                        del self.old_names[msg['index']]
                         del self[self[msg['index']]['ifname']]
                         del self[msg['index']]
                 elif msg.get('event', None) == 'RTM_NEWADDR':
