@@ -1,4 +1,5 @@
 
+from socket import htons
 from socket import AF_INET
 from socket import AF_INET6
 from socket import AF_UNSPEC
@@ -10,6 +11,11 @@ from pyroute2.netlink import NLM_F_CREATE
 from pyroute2.netlink import NLM_F_EXCL
 from pyroute2.netlink import marshal
 from pyroute2.netlink.generic import NETLINK_ROUTE
+from pyroute2.netlink.rtnl.tcmsg import tcmsg
+from pyroute2.netlink.rtnl.tcmsg import get_htb_parameters
+from pyroute2.netlink.rtnl.tcmsg import get_tbf_parameters
+from pyroute2.netlink.rtnl.tcmsg import get_sfq_parameters
+from pyroute2.netlink.rtnl.tcmsg import get_u32_parameters
 from pyroute2.netlink.rtnl.rtmsg import rtmsg
 from pyroute2.netlink.rtnl.ndmsg import ndmsg
 from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
@@ -78,6 +84,9 @@ RTM_GETNEIGHTBL = 66
 RTM_SETNEIGHTBL = 67
 (RTM_NAMES, RTM_VALUES) = map_namespace('RTM', globals())
 
+TC_H_INGRESS = 0xfffffff1
+TC_H_ROOT = 0xffffffff
+
 rtypes = {'RTN_UNSPEC': 0,
           'RTN_UNICAST': 1,      # Gateway or direct route
           'RTN_LOCAL': 2,        # Accept locally
@@ -127,7 +136,13 @@ class marshal_rtnl(marshal):
                RTM_NEWROUTE: rtmsg,
                RTM_DELROUTE: rtmsg,
                RTM_NEWNEIGH: ndmsg,
-               RTM_DELNEIGH: ndmsg}
+               RTM_DELNEIGH: ndmsg,
+               RTM_NEWQDISC: tcmsg,
+               RTM_DELQDISC: tcmsg,
+               RTM_NEWTCLASS: tcmsg,
+               RTM_DELTCLASS: tcmsg,
+               RTM_NEWTFILTER: tcmsg,
+               RTM_DELTFILTER: tcmsg}
 
     def fix_message(self, msg):
         try:
@@ -144,12 +159,32 @@ class iproute(netlink):
         RTNLGRP_IPV4_ROUTE |\
         RTNLGRP_IPV6_ROUTE |\
         RTNLGRP_NEIGH |\
-        RTNLGRP_LINK
+        RTNLGRP_LINK |\
+        RTNLGRP_TC
 
     # 8<---------------------------------------------------------------
     #
     # Listing methods
     #
+    def get_qdiscs(self, family=AF_UNSPEC):
+        msg = tcmsg()
+        msg['family'] = family
+        return self.nlm_request(msg, RTM_GETQDISC)
+
+    def get_filters(self, family=AF_UNSPEC, index=0, handle=0, parent=0):
+        msg = tcmsg()
+        msg['family'] = family
+        msg['index'] = index
+        msg['handle'] = handle
+        msg['parent'] = parent
+        return self.nlm_request(msg, RTM_GETTFILTER)
+
+    def get_classes(self, family=AF_UNSPEC, index=0):
+        msg = tcmsg()
+        msg['family'] = family
+        msg['index'] = index
+        return self.nlm_request(msg, RTM_GETTCLASS)
+
     def get_links(self, links=None, family=AF_UNSPEC):
         '''
         Get network interfaces sepcifications.
@@ -383,7 +418,7 @@ class iproute(netlink):
             nla = key.upper()
             if not nla.startswith('IFLA_'):
                 nla = 'IFLA_%s' % (nla)
-            msg['attrs'].append((nla, kwarg[key]))
+            msg['attrs'].append([nla, kwarg[key]])
 
         return self.nlm_request(msg, msg_type=action, msg_flags=msg_flags)
 
@@ -415,10 +450,47 @@ class iproute(netlink):
         msg['prefixlen'] = mask
         msg['scope'] = 0xfe
         if family == AF_INET:
-            msg['attrs'] = (('IFA_LOCAL', address),
-                            ('IFA_ADDRESS', address))
+            msg['attrs'] = [['IFA_LOCAL', address],
+                            ['IFA_ADDRESS', address]]
         elif family == AF_INET6:
-            msg['attrs'] = (('IFA_ADDRESS', address), )
+            msg['attrs'] = [['IFA_ADDRESS', address]]
+        return self.nlm_request(msg, msg_type=action, msg_flags=flags)
+
+    def tc(self, action, kind, interface, handle, **kwarg):
+        '''
+        '''
+
+        flags = NLM_F_REQUEST | NLM_F_ACK | NLM_F_CREATE | NLM_F_EXCL
+        msg = tcmsg()
+        msg['index'] = interface
+        msg['handle'] = handle
+        opts = None
+        if kind == 'ingress':
+            msg['parent'] = TC_H_INGRESS
+        elif kind == 'tbf':
+            msg['parent'] = TC_H_ROOT
+            if kwarg:
+                # kwarg is empty for delete
+                opts = get_tbf_parameters(kwarg)
+        elif kind == 'htb':
+            msg['parent'] = kwarg.get('parent', TC_H_ROOT)
+            if kwarg:
+                # kwarg is empty for delete
+                opts = get_htb_parameters(kwarg)
+        elif kind == 'sfq':
+            msg['parent'] = kwarg.get('parent', TC_H_ROOT)
+            if kwarg:
+                # kwarg is empty for delete
+                opts = get_sfq_parameters(kwarg)
+        elif kind == 'u32':
+            msg['parent'] = kwarg.get('parent')
+            msg['info'] = htons(kwarg.get('protocol', 0) & 0xffff) |\
+                ((kwarg.get('prio', 0) << 16) & 0xffff0000)
+            if kwarg:
+                # kwarg is empty for delete
+                opts = get_u32_parameters(kwarg)
+        msg['attrs'] = [['TCA_KIND', kind],
+                        ['TCA_OPTIONS', opts]]
         return self.nlm_request(msg, msg_type=action, msg_flags=flags)
 
     def route(self, action, prefix, mask, table=254,
@@ -456,12 +528,12 @@ class iproute(netlink):
         msg['type'] = rtypes[rtype]
         msg['scope'] = rtscopes[rtscope]
         msg['dst_len'] = mask
-        msg['attrs'] = [('RTA_DST', prefix),
-                        ('RTA_TABLE', table)]
+        msg['attrs'] = [['RTA_DST', prefix],
+                        ['RTA_TABLE', table]]
         if interface is not None:
-            msg['attrs'].append(('RTA_OIF', interface))
+            msg['attrs'].append(['RTA_OIF', interface])
         if gateway is not None:
-            msg['attrs'].append(('RTA_GATEWAY', gateway))
+            msg['attrs'].append(['RTA_GATEWAY', gateway])
 
         return self.nlm_request(msg, msg_type=action,
                                 msg_flags=flags)
